/*
  ==============================================================================

   This file is part of the JUCE library - "Jules' Utility Class Extensions"
   Copyright 2004 by Julian Storer.

  ------------------------------------------------------------------------------

   JUCE can be redistributed and/or modified under the terms of the
   GNU General Public License, as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later version.

   JUCE is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JUCE; if not, visit www.gnu.org/licenses or write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
   Boston, MA 02111-1307 USA

  ------------------------------------------------------------------------------

   If you'd like to release a closed-source product which uses JUCE, commercial
   licenses are also available: visit www.rawmaterialsoftware.com/juce for
   more information.

  ==============================================================================
*/

/*
  ==============================================================================
  VST-Sepcific Template modifications by ModuLR
  ==============================================================================
*/

#ifdef _MSC_VER
#pragma warning (disable : 4996)
#endif

#include "audioeffectx.h"
#include "AEffEditor.hpp"
#include "juce.h"
#include "JuceVstParams.h"
#include "JuceVstEditor.h"

#ifdef JUCE_WIN32
  #undef Rectangle
  #define Array tempArray
  #include <windows.h>
  #undef Array
#else
  #include <Carbon/Carbon.h>
#endif

#ifdef JUCE_MAC
class CheckVisibilityTimer  : public Timer
{
    WindowRef hostWindow;
    Component* pluginWindow;

public:
    CheckVisibilityTimer (WindowRef hostWindow_, 
                          Component* const pluginWindow_)
        : hostWindow (hostWindow_),
          pluginWindow (pluginWindow_)
    {
        startTimer (20);
    }

    void timerCallback()
    {
        pluginWindow->setVisible (IsWindowVisible (hostWindow));
    }
};
#endif


// There is a bug in the mac version such that when numInstances == 0, the
// first instance of the editorComp->getWindowHandle() will fail to supply
// a pointer for the PluginWindow.  All instances after will be ok, until
// it's back down to the one.  I set it to 1 temporary, so I can at least
// test single instances of the plugin.
static int numInstances = 1;

//===============================================================================
// MyJuceVST is wrapped as AudioEffectX
//===============================================================================
class MyJuceVST       : public AudioEffectX,
                        public Timer,
                        public ComponentListener,
						public ActionListener
{

//------------------------------------------------------------------------------

private:  
	double                   sampleRate;
    int                      bufferSize;
    MemoryBlock              chunkMemory;
    EditorComponent*         editorComp;
    ERect                    editorSize;
    VstEvents*               outgoingEvents;
    int                      outgoingEventSize;
    OwnedArray<MidiMessage>  midiEvents;



    void ensureOutgoingEventSize (int numEvents)
    {
        if (outgoingEventSize < numEvents)
        {
            numEvents += 32;
            const int size = 16 + sizeof (VstEvent*) * numEvents;

            if (outgoingEvents == 0)
                outgoingEvents = (VstEvents*) juce_calloc (size);
            else
                outgoingEvents = (VstEvents*) juce_realloc (outgoingEvents, size);

            for (int i = outgoingEventSize; i < numEvents; ++i)
            {
                VstMidiEvent* const e = (VstMidiEvent*) juce_calloc (sizeof (VstMidiEvent));
                e->type = kVstMidiType;
                e->byteSize = 24;

                outgoingEvents->events[i] = (VstEvent*) e;
            }

            outgoingEventSize = numEvents;
        }
    }

#ifdef JUCE_MAC
    WindowRef hostWindow;
    WindowGroupRef group;
    CheckVisibilityTimer* checkVisibilityTimer;
#else
    HWND hostWindow;
#endif

//------------------------------------------------------------------------------

public:

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Basic structure for Programs
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	struct MyJuceVSTPrograms
	{
		float param[NUM_PARAMS];
        char  name[50];
	};
	
	// create program array
	MyJuceVSTPrograms programs[NUM_PROGRAMS];
	
	// create parameter array
	float param[NUM_PARAMS];



	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Chunk data structure - e.g., store all the program/parameter info here, and 
	// pretty much whatever other data you want retreivable
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	struct chunkData
	{
		MyJuceVSTPrograms programs[NUM_PROGRAMS];
	};
	chunkData chunk;


	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Constructor
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	MyJuceVST (audioMasterCallback audioMaster) : AudioEffectX (audioMaster,NUM_PROGRAMS,NUM_PARAMS)
    {
        ++numInstances;

        sampleRate        = 44100.0;
        bufferSize        = 512;
        outgoingEvents    = 0;
        outgoingEventSize = 0;
        editorComp        = 0;


#ifdef JUCE_MAC
        hostWindow = 0;
        checkVisibilityTimer = 0;
        group = 0;
#endif
	
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		// Basic configuration sent from plugin to host during creation of the effect/instrument
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        setUniqueID                 (FILTER_ID);
        setNumInputs                (NUM_INPUTS);
        setNumOutputs               (NUM_OUTPUTS);
        setInitialDelay             (LATENCY);
		wantEvents                  (HAS_MIDI_IN);
		programsAreChunks           (USE_CHUNKS);
		canMono                     (CAN_MONO);
		canProcessReplacing         (CAN_PROCREPLACING);

		if (HAS_EDITOR) 
		{ 
			cEffect.flags |= effFlagsHasEditor;
			createEditorComp();
		}


		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		// Inititalize Programs
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		for (int i = 0; i < NUM_PROGRAMS; i++)
		{
			strcpy(programs[i].name,"Preset");
			
			for (int j = 0; j < NUM_PARAMS; j++)
			{ 
				programs[i].param[j] = 0;
			}
		}

		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		// Initialiaze your default parameters here
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//setParameterAutomated                (<insertparam>,0.5);
    }




	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Deconstructor
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	~MyJuceVST()
    {
#ifdef JUCE_MAC
        deleteAndZero (checkVisibilityTimer);
#endif

		if (editorComp != 0)
            deleteEditor();

        if (outgoingEvents != 0)
        {
            for (int i = outgoingEventSize; --i >= 0;)
                juce_free (outgoingEvents->events[i]);

            juce_free (outgoingEvents);
        }

        jassert (editorComp == 0);

        if (--numInstances == 0)
        {
#ifdef JUCE_MAC
            DeletedAtShutdown::deleteAll();
            Thread::stopAllThreads (3000);
#endif
        }
    }


	
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Send host effect name, vendor info, and if plugin is an effect or instrument
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	bool getEffectName    (char* name) { name=(char*)FILTER_NAME; return true;}
	bool getVendorString  (char* text) { text=(char*)VENDOR_NAME; return true;}
	bool getProductString (char* text) { text=(char*)FILTER_NAME; return true;}
	long getVendorVersion ()	       { return 1000;                         }
	VstPlugCategory getPlugCategory()  { return IS_A_SYNTH ? kPlugCategSynth : kPlugCategEffect; }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//Set sample rate and block size
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void setSampleRate (float newRate) { sampleRate = newRate; }
	void setBlockSize  (long newSize)  { bufferSize = newSize; }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//Return CanDo's to host
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	long canDo (char* text)
	{	
		if (text == "receiveVstEvents")                      return 1;
		if (text == "sendVstEvents")                         return 1;
		if ((text == "receiveVstMidiEvents") && HAS_MIDI_IN) return 1;
		if ((text == "sendVstMidiEvent") && HAS_MIDI_OUT)    return 1;
		if ((text == "receiveVstTimeInfo") && SYNC_TO_HOST)  return 1;
		
		return -1;
	}




//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Programs
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Set Programs
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	void setProgram (long program)
    {
		if ((program < 0) || (program > NUM_PROGRAMS-1)) 
		{

		}
		else
		{
			for (int i = 0; i < NUM_PARAMS-1; i++)
			{
				curProgram = program;
				setParameter(i,programs[curProgram].param[i]);
			}
		}
	}
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Set ProgramName
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	void setProgramName (char* name)
    {
		strcpy(programs[curProgram].name, name);
    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Get ProgramName
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	void getProgramName (char* name)
    {
		strcpy(name, programs[curProgram].name);
    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Get ProgramName based on index
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	bool getProgramNameIndexed (long category, long index, char* text)
    {
		if ((index >= 0) && (index < NUM_PROGRAMS)) 
        {
			strcpy(text,programs[index].name);
            return true;
        }

        return false;
    }




//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Parameters
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Get Parameter value
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	float getParameter (long index)
    {
		if (index >= NUM_PARAMS)
		{
			return 0;
		}
		else 
		{
			return param[index];
		}
    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Set Parameter Value including inside the editor (if using one)
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	void setParameter (long index, float value)
    {
		param[index] = value;
        programs[curProgram].param[index] = value;

		// editorComp->setEditorParameter(index, value);
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Callback from editor telling you that a parameter is manually being
	// changed, and thus wee need to update our param array values
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void actionListenerCallback (const String& m)
	{
	}


	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Parameter Name(s) displayed by the Host
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void getParameterName (long index, char* text)
    {
		switch (index)
		{		
			default:   strcpy(text,"");        break;
		}

    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Display in GUI'less VSTs by the Host
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    void getParameterDisplay (long index, char* text)
    {
		switch (index)
		{
		default: strcpy(text,""); break;
		}
    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Get Parameter Label which is added behind the value displayed by the host (eg, "<value> %"
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void getParameterLabel (long index, char* label)
    {
		switch (index)
		{
			default: strcpy(label,""); break;
		}
    }




//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Chunks can be used to store/recall parameter info in/from a single chunk of memmory. 
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//getChunk - Hosts calls function and expects a pointer to a structure which
	//           contains your stored data.  This is a functional (tho not and
	//           ideal) example of that.
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	long getChunk (void** data, bool isPreset)
    {
		//chunkMemory.setSize (0);
		//
		//// store current program/parameter info into chunk
		//for (int i=0; i < NUM_PROGRAMS; i++)
		//{
		//	strcpy(chunk.programs[i].name,programs[i].name);
		//	
		//	for (int j=0; j < NUM_PARAMS; j++)
		//	{
		//		chunk.programs[i].param[j] = programs[i].param[j];
		//	}
		//}
		//	
		//// place the content in a MemoryBlock.
		//chunkMemory.append(&chunk,sizeof(chunk));

        chunkMemory.setSize (0);
        editorComp->getStateInformation (chunkMemory);
        *data = (void*) chunkMemory;
		
		
		//*data = (void*)chunkMemory;
	
        // because the chunk is only needed temporarily by the host (or at least you'd
        // hope so) we'll give it a while and then free it. As long as this is called on
        // the message thread there's no danger of the timer going off too soon.
        
		startTimer (2000);

        return chunkMemory.getSize();
    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//setChunk - Hosts calls function and provides a pointer which must be 
 	//           typecast to original data structure.
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	long setChunk (void* data, long byteSize, bool isPreset)
    {
		editorComp->setStateInformation(data,byteSize);
		return 0;
    }

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//Timer callback will free MemoryBlock used for chunks          
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    void timerCallback()
    {
        stopTimer();
        chunkMemory.setSize (0);
    }




//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Setup input/output channels + labeling
// (eg, shown when hovering over input/output inside Tracktion's Rack Filterss)
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Input Labels
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	bool getInputProperties(long index, VstPinProperties* properties)
	{
		if (index < NUM_INPUTS)
		{
			switch (index)
			{
				default : strcpy(properties->label,"Input");
			}
			properties->flags = kVstPinIsActive;
			properties->flags = properties->flags | kVstPinIsStereo;
			return true;
		}
		return false;
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Output Labels
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	bool getOutputProperties(long index, VstPinProperties* properties)
	{

		if (index < NUM_OUTPUTS)
		{
			switch (index)
			{
				default : strcpy(properties->label,"Output");
			}
			properties->flags = kVstPinIsActive;
			properties->flags = properties->flags | kVstPinIsStereo;
			return true;
		}
		return false;
	}





//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Midi Event Handler - places into incomming midi events into a JUCEY event list.
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	long processEvents (VstEvents* events)
    {
        if (HAS_MIDI_IN)
        {
            for (int i = 0; i < events->numEvents; ++i)
            {
                const VstEvent* const e = events->events[i];

                if (e != 0 && e->type == kVstMidiType)
                {
                    const VstMidiEvent* const vme = (const VstMidiEvent*) e;

                    int numUsed = 0;
                    midiEvents.add (new MidiMessage ((const uint8*) vme->midiData, 4, 
                                                     numUsed, 0, vme->deltaFrames));

                    if (numUsed == 0)
                        midiEvents.removeLast();
                }
            }
            return 1;
        }
        else
        {
            return 0;
        }
    }

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Process Outgoing Midi Events
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void processMidi(int numSamplesToProcess)        
	{	
		if  ( (! HAS_MIDI_IN) && (! IS_MIDIONLY) )
            midiEvents.clear();

        if (HAS_MIDI_OUT)
        {
            ensureOutgoingEventSize (midiEvents.size());
            outgoingEvents->numEvents = midiEvents.size();

            for (int i = 0; i < midiEvents.size(); ++i)
            {
                const MidiMessage* const m = midiEvents.getUnchecked(i);

                VstMidiEvent* const vme = (VstMidiEvent*) outgoingEvents->events[i];

                vme->midiData[0] = m->getRawData()[0];
                vme->midiData[1] = m->getRawData()[1];
                vme->midiData[2] = m->getRawData()[2];
                vme->midiData[3] = m->getRawData()[3];
                vme->deltaFrames = roundDoubleToInt (m->getTimeStamp());

                jassert (vme->deltaFrames >= 0 && vme->deltaFrames < numSamplesToProcess);
            }

            sendVstEventsToHost (outgoingEvents);

            midiEvents.clear();
        }
	}




//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Audio Processing Routines
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    void process (float** inputs, float** outputs, const long num, const bool accumulate)
    {
        const AudioSampleBuffer input (inputs, NUM_INPUTS, num);
        AudioSampleBuffer output (outputs, NUM_OUTPUTS, num);

        //filter->processBlock (input, output, accumulate, midiEvents);
    }


	//---------------------------------------------------------------------------
	// Accumulating Process Function
	//---------------------------------------------------------------------------
	void process (float** inputs, float** outputs, long numSamplesToProcess)
    {

	    float *in1  =  inputs[0];
	    float *in2  =  inputs[1];
	    float *out1 = outputs[0];
	    float *out2 = outputs[1];

	    while (--numSamplesToProcess >= 0)
	    {
			(*out1++) += (*in1++);
			(*out2++) += (*in2++);
		}

        
		//process (inputs, outputs, num, true);
    }

	//---------------------------------------------------------------------------
	// Replacing PRocess Function
	//---------------------------------------------------------------------------
	void processReplacing (float** inputs, float** outputs, long numSamplesToProcess)
    {
		processMidi(numSamplesToProcess);

	    float *in1  =  inputs[0];
	    float *in2  =  inputs[1];
	    float *out1 = outputs[0];
	    float *out2 = outputs[1];

	    while (--numSamplesToProcess >= 0)
	    {
			(*out1++) = (*in1++);
			(*out2++) = (*in2++);
		}		

        //process (inputs, outputs, num, false);
    }


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Suspend and Resume Routines.. you can flush buffers and what not here.
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	void resume()
    {
        midiEvents.clear();
        
		sampleRate = getSampleRate();
        bufferSize = getBlockSize();
        
		AudioEffectX::resume();
		
		wantEvents(HAS_MIDI_IN);
		
		if (HAS_MIDI_OUT)
            ensureOutgoingEventSize (64);

    }

	//---------------------------------------------------------------------------

    void suspend()
    {
    	AudioEffectX::suspend();
        midiEvents.clear();
    }




//===============================================================================
// Editor (GUI)
//===============================================================================

	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Create the Editor Component
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void createEditorComp()
    {
        /*if (numInstances == 1)
		{
			editorComp = new EditorComponent();
			deleteAndZero(editorComp);
		}*/
		
		if (editorComp == 0)
        {
            editorComp = new EditorComponent();

        }
    }


	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Delete the Editor Component
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    void deleteEditor()
    {
#ifdef JUCE_MAC
        deleteAndZero (checkVisibilityTimer);
        
        if (hostWindow != 0)
        {
            SetWindowGroup (hostWindow, 0);
            hostWindow = 0;
        }
        
        if (editorComp != 0)
            SetWindowGroup ((WindowRef) editorComp->getWindowHandle(), 0);
        
        if (group != 0)
            ReleaseWindowGroup (group);
#endif       
        deleteAndZero (editorComp);		
    }


	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Open Editor
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	void openEditor(void* ptr)
	{
		if (editorComp != 0)
        {

			editorComp->setOpaque (true);
            editorComp->setVisible (false);
			editorComp->addToDesktop (0);


#ifdef JUCE_WIN32
            hostWindow = (HWND) ptr;
            HWND editorWnd = (HWND) editorComp->getWindowHandle();

            SetParent (editorWnd, hostWindow);

            DWORD val = GetWindowLong (editorWnd, GWL_STYLE);
            val = (val & ~WS_POPUP) | WS_CHILD;
            SetWindowLong (editorWnd, GWL_STYLE, val);

#else

            WindowRef plugWindow = (WindowRef) editorComp->getWindowHandle();
            hostWindow = (WindowRef) ptr;
            WindowGroupRef hostParentGroup = GetWindowGroup (hostWindow);
            ChangeWindowAttributes (plugWindow, kWindowNoShadowAttribute, 0);

            Rect r;
            r.left = editorComp->getX();
            r.right = editorComp->getRight();
            r.top = editorComp->getY();
            r.bottom = editorComp->getBottom();
            SetWindowBounds (hostWindow, kWindowContentRgn, &r);

            CreateWindowGroup (0, &group);
            SetWindowGroupParent (hostParentGroup, group);
            SetWindowGroup (hostWindow, group);
            SetWindowGroup (plugWindow, group);
            BringToFront (plugWindow);
            ChangeWindowGroupAttributes (group, 
                                             kWindowGroupAttrSelectAsLayer
                                              | kWindowGroupAttrLayerTogether
                                              | kWindowGroupAttrMoveTogether
                                              | kWindowGroupAttrSharedActivation
                                              | kWindowGroupAttrHideOnCollapse,
                                             0);
            
            checkVisibilityTimer = new CheckVisibilityTimer (hostWindow, editorComp);
#endif
            
			editorComp->setVisible (true);
            editorComp->addComponentListener (this);
			
		    // add an Action Listener.. so you can report back 
			// parameter changes to the callback.
			editorComp->addActionListener(this);
		}
	}

	void idleEditor()
	{
		
	}
	
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Close Editor
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	void closeEditor()
	{
		editorComp->removeAllActionListeners();

#ifdef JUCE_WIN32
        editorComp->removeFromDesktop();
#endif

#ifdef JUCE_MAC
        deleteAndZero (checkVisibilityTimer);
        
        if (hostWindow != 0)
        {
            SetWindowGroup (hostWindow, 0);
            hostWindow = 0;
        }
        
        if (editorComp != 0)
            SetWindowGroup ((WindowRef) editorComp->getWindowHandle(), 0);
        
        if (group != 0)
            ReleaseWindowGroup (group);
#endif
    }



	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Get Editor Rect (Interface window sizing sent to Host)
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	long getEditorRect(void* ptr)
	{
		if (editorComp != 0)
        {
	        editorSize.left = 0;
            editorSize.top = 0;
            editorSize.right = editorComp->getWidth();
            editorSize.bottom = editorComp->getHeight();

            *((ERect**) ptr) = &editorSize;

            return (long) &editorSize;
         }
         else
         {
			return 0;
         }
	}



	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Handle Editor Resizing
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    void componentMovedOrResized (Component& component,
                                  bool wasMoved,
                                  bool wasResized)
    {
		if (editorComp != 0)
		{ 
			editorSize.left   = 0;			
			editorSize.top    = 0;
			editorSize.right  = editorComp->getWidth();
			editorSize.bottom = editorComp->getHeight();

#ifdef JUCE_MAC
            Rect r;
            r.left = editorComp->getX();
            r.right = editorComp->getRight();
            r.top = editorComp->getY();
            r.bottom = editorComp->getBottom();
            SetWindowBounds (hostWindow, kWindowContentRgn, &r);

#else
            SetWindowPos (hostWindow, 0,
                          0, 0, editorSize.right, editorSize.bottom,
                          SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER);

            if (GetParent (hostWindow) != 0)
                SetWindowPos (GetParent (hostWindow), 0,
                              0, 0, editorSize.right, editorSize.bottom,
                              SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER);
#endif
			AudioEffectX::sizeWindow (editorSize.right, editorSize.bottom);
		}
    }
    
	
	
	long dispatcher (long opCode, long index, long value, void *ptr, float opt)
    {
		switch (opCode)
		{
			case effEditOpen:		{ openEditor(ptr); return 0;	break; }
			case effEditClose:		{ closeEditor(); return 0;		break; }
			case effEditGetRect:	{ return getEditorRect(ptr);	break; }
			case effEditIdle:		{ idleEditor(); return 0;		break; }
			case effKeysRequired:   { return 0;                     break; }
		
			default: { return AudioEffectX::dispatcher (opCode, index, value, ptr, opt); break; }
		}
    }
};







//==============================================================================
#ifdef JUCE_MAC
extern "C" AEffect* main_macho (audioMasterCallback audioMaster)
{
    static bool initialised = false;
    if (! initialised)
    {
        initialised = true;
        SystemStats::initialiseStats();
        MessageManager::getInstance();
    }

#else
__declspec(dllexport) void* main (audioMasterCallback audioMaster)
{
#endif
    try
    {
	    if (audioMaster (0, audioMasterVersion, 0, 0, 0, 0) != 0)
        {
	        MyJuceVST* const effect = new MyJuceVST (audioMaster);
	        return effect->getAeffect();
        }
    }
    catch (...)
    {}

    return 0;
}

//==============================================================================
#ifdef JUCE_WIN32
    BOOL WINAPI DllMain (HINSTANCE instance, DWORD dwReason, LPVOID)
    {
        if (dwReason == DLL_PROCESS_ATTACH)
        {
            PlatformUtilities::setCurrentModuleInstanceHandle (instance);
            SystemStats::initialiseStats();
            MessageManager::getInstance();
        }
        else if (dwReason == DLL_PROCESS_DETACH)
        {
            DeletedAtShutdown::deleteAll();
            Thread::stopAllThreads (3000);
        }

        return TRUE;
    }
#endif
